@@ -61,11 +61,15 @@
 // HT_EMPTY_FLT          -- how empty before we halve size
 // HT_MIN_BUCKETS        -- default smallest bucket size
 //
+// You can also change enlarge_resize_percent (which defaults to
+// HT_OCCUPANCY_FLT), and shrink_resize_percent (which defaults to
+// HT_EMPTY_FLT) with set_resizing_parameters().
+//
 // How to decide what values to use?
-// HT_EMPTY_FLT's default of .4 * OCCUPANCY_FLT, is probably good.
+// shrink_resize_percent's default of .4 * OCCUPANCY_FLT, is probably good.
 // HT_MIN_BUCKETS is probably unnecessary since you can specify
 // (indirectly) the starting number of buckets at construct-time.
-// For HT_OCCUPANCY_FLT, you can use this chart to try to trade-off
+// For enlarge_resize_percent, you can use this chart to try to trade-off
 // expected lookup time to the space taken up.  By default, this
 // code uses quadratic probing, though you can change it to linear
 // via _JUMP below if you really want to.
@@ -75,7 +79,7 @@
 // Quadratic collision resolution   1 - ln(1-L) - L/2    1/(1-L) - L - ln(1-L)
 // Linear collision resolution     [1+1/(1-L)]/2         [1+1/(1-L)2]/2
 //
-// -- HT_OCCUPANCY_FLT --         0.10  0.50  0.60  0.75  0.80  0.90  0.99
+// -- enlarge_resize_percent --         0.10  0.50  0.60  0.75  0.80  0.90  0.99
 // QUADRATIC COLLISION RES.
 //    probes/successful lookup    1.05  1.44  1.62  2.01  2.21  2.85  5.11
 //    probes/unsuccessful lookup  1.11  2.19  2.82  4.64  5.81  11.4  103.6
@@ -93,13 +97,12 @@
 #define JUMP_(key, num_probes)    ( num_probes )
 
 
-// Hashtable class, used to implement the hashed associative containers
-// hash_set and hash_map.
-
 #include <google/sparsehash/sparseconfig.h>
 #include <assert.h>
+#include <stdio.h>
 #include <stdlib.h>             // for abort()
 #include <algorithm>            // For swap(), eg
+#include <stdexcept>            // For length_error
 #include <iostream>             // For cerr
 #include <memory>               // For uninitialized_fill, uninitialized_copy
 #include <utility>              // for pair<>
@@ -110,22 +113,37 @@ _START_GOOGLE_NAMESPACE_
 
 using STL_NAMESPACE::pair;
 
+// Hashtable class, used to implement the hashed associative containers
+// hash_set and hash_map.
+
+// Value: what is stored in the table (each bucket is a Value).
+// Key: something in a 1-to-1 correspondence to a Value, that can be used
+//      to search for a Value in the table (find() takes a Key).
+// HashFcn: Takes a Key and returns an integer, the more unique the better.
+// ExtractKey: given a Value, returns the unique Key associated with it.
+// SetKey: given a Value* and a Key, modifies the value such that
+//         ExtractKey(value) == key.  We guarantee this is only called
+//         with key == deleted_key or key == empty_key.
+// EqualKey: Given two Keys, says whether they are the same (that is,
+//           if they are both associated with the same Value).
+// Alloc: STL allocator to use to allocate memory.  Currently ignored.
+
 template <class Value, class Key, class HashFcn,
-          class ExtractKey, class EqualKey, class Alloc>
+          class ExtractKey, class SetKey, class EqualKey, class Alloc>
 class dense_hashtable;
 
-template <class V, class K, class HF, class ExK, class EqK, class A>
+template <class V, class K, class HF, class ExK, class SetK, class EqK, class A>
 struct dense_hashtable_iterator;
 
-template <class V, class K, class HF, class ExK, class EqK, class A>
+template <class V, class K, class HF, class ExK, class SetK, class EqK, class A>
 struct dense_hashtable_const_iterator;
 
 // We're just an array, but we need to skip over empty and deleted elements
-template <class V, class K, class HF, class ExK, class EqK, class A>
+template <class V, class K, class HF, class ExK, class SetK, class EqK, class A>
 struct dense_hashtable_iterator {
  public:
-  typedef dense_hashtable_iterator<V,K,HF,ExK,EqK,A>       iterator;
-  typedef dense_hashtable_const_iterator<V,K,HF,ExK,EqK,A> const_iterator;
+  typedef dense_hashtable_iterator<V,K,HF,ExK,SetK,EqK,A>       iterator;
+  typedef dense_hashtable_const_iterator<V,K,HF,ExK,SetK,EqK,A> const_iterator;
 
   typedef STL_NAMESPACE::forward_iterator_tag iterator_category;
   typedef V value_type;
@@ -135,7 +153,7 @@ struct dense_hashtable_iterator {
   typedef V* pointer;
 
   // "Real" constructor and default constructor
-  dense_hashtable_iterator(const dense_hashtable<V,K,HF,ExK,EqK,A> *h,
+  dense_hashtable_iterator(const dense_hashtable<V,K,HF,ExK,SetK,EqK,A> *h,
                            pointer it, pointer it_end, bool advance)
     : ht(h), pos(it), end(it_end)   {
     if (advance)  advance_past_empty_and_deleted();
@@ -165,17 +183,17 @@ struct dense_hashtable_iterator {
 
 
   // The actual data
-  const dense_hashtable<V,K,HF,ExK,EqK,A> *ht;
+  const dense_hashtable<V,K,HF,ExK,SetK,EqK,A> *ht;
   pointer pos, end;
 };
 
 
 // Now do it all again, but with const-ness!
-template <class V, class K, class HF, class ExK, class EqK, class A>
+template <class V, class K, class HF, class ExK, class SetK, class EqK, class A>
 struct dense_hashtable_const_iterator {
  public:
-  typedef dense_hashtable_iterator<V,K,HF,ExK,EqK,A>       iterator;
-  typedef dense_hashtable_const_iterator<V,K,HF,ExK,EqK,A> const_iterator;
+  typedef dense_hashtable_iterator<V,K,HF,ExK,SetK,EqK,A>       iterator;
+  typedef dense_hashtable_const_iterator<V,K,HF,ExK,SetK,EqK,A> const_iterator;
 
   typedef STL_NAMESPACE::forward_iterator_tag iterator_category;
   typedef V value_type;
@@ -185,8 +203,9 @@ struct dense_hashtable_const_iterator {
   typedef const V* pointer;
 
   // "Real" constructor and default constructor
-  dense_hashtable_const_iterator(const dense_hashtable<V,K,HF,ExK,EqK,A> *h,
-                                 pointer it, pointer it_end, bool advance)
+  dense_hashtable_const_iterator(
+      const dense_hashtable<V,K,HF,ExK,SetK,EqK,A> *h,
+      pointer it, pointer it_end, bool advance)
     : ht(h), pos(it), end(it_end)   {
     if (advance)  advance_past_empty_and_deleted();
   }
@@ -218,12 +237,12 @@ struct dense_hashtable_const_iterator {
 
 
   // The actual data
-  const dense_hashtable<V,K,HF,ExK,EqK,A> *ht;
+  const dense_hashtable<V,K,HF,ExK,SetK,EqK,A> *ht;
   pointer pos, end;
 };
 
 template <class Value, class Key, class HashFcn,
-          class ExtractKey, class EqualKey, class Alloc>
+          class ExtractKey, class SetKey, class EqualKey, class Alloc>
 class dense_hashtable {
  public:
   typedef Key key_type;
@@ -238,18 +257,23 @@ class dense_hashtable {
   typedef value_type&       reference;
   typedef const value_type& const_reference;
   typedef dense_hashtable_iterator<Value, Key, HashFcn,
-                                   ExtractKey, EqualKey, Alloc>
+                                   ExtractKey, SetKey, EqualKey, Alloc>
   iterator;
 
   typedef dense_hashtable_const_iterator<Value, Key, HashFcn,
-                                          ExtractKey, EqualKey, Alloc>
+                                         ExtractKey, SetKey, EqualKey, Alloc>
   const_iterator;
 
-  // How full we let the table get before we resize.  Knuth says .8 is
-  // good -- higher causes us to probe too much, though saves memory
-  static const float HT_OCCUPANCY_FLT; // = 0.8;
+  // These come from tr1.  For us they're the same as regular iterators.
+  typedef iterator local_iterator;
+  typedef const_iterator const_local_iterator;
 
-  // How empty we let the table get before we resize lower.
+  // How full we let the table get before we resize, by default.
+  // Knuth says .8 is good -- higher causes us to probe too much,
+  // though it saves memory.
+  static const float HT_OCCUPANCY_FLT; // = 0.5;
+
+  // How empty we let the table get before we resize lower, by default.
   // (0.0 means never resize lower.)
   // It should be less than OCCUPANCY_FLT / 2 or we thrash resizing
   static const float HT_EMPTY_FLT; // = 0.4 * HT_OCCUPANCY_FLT
@@ -258,7 +282,12 @@ class dense_hashtable {
   // Must be a power of two, and at least 4.
   // Note, however, that for a given hashtable, the initial size is a
   // function of the first constructor arg, and may be >HT_MIN_BUCKETS.
-  static const size_t HT_MIN_BUCKETS  = 32;
+  static const size_t HT_MIN_BUCKETS = 4;
+
+  // By default, if you don't specify a hashtable size at
+  // construction-time, we use this size.  Must be a power of two, and
+  // at least HT_MIN_BUCKETS.
+  static const size_t HT_DEFAULT_STARTING_BUCKETS = 32;
 
 
   // ITERATOR FUNCTIONS
@@ -271,15 +300,42 @@ class dense_hashtable {
   const_iterator end() const   { return const_iterator(this, table + num_buckets,
                                                        table+num_buckets,true);}
 
+  // These come from tr1 unordered_map.  They iterate over 'bucket' n.
+  // For sparsehashtable, we could consider each 'group' to be a bucket,
+  // I guess, but I don't really see the point.  We'll just consider
+  // bucket n to be the n-th element of the sparsetable, if it's occupied,
+  // or some empty element, otherwise.
+  local_iterator begin(size_type i) {
+    return local_iterator(this, table + i, table + i+1, false);
+  }
+  local_iterator end(size_type i) {
+    local_iterator it = begin(i);
+    if (!test_empty(i) && !test_deleted(i))
+      ++it;
+    return it;
+  }
+  const_local_iterator begin(size_type i) const {
+    return const_local_iterator(this, table + i, table + i+1, false);
+  }
+  const_local_iterator end(size_type i) const {
+    const_local_iterator it = begin(i);
+    if (!test_empty(i) && !test_deleted(i))
+      ++it;
+    return it;
+  }
+
   // ACCESSOR FUNCTIONS for the things we templatize on, basically
   hasher hash_funct() const { return hash; }
   key_equal key_eq() const  { return equals; }
 
+  // Accessor function for statistics gathering.
+  int num_table_copies() const { return num_ht_copies; }
+
+ private:
   // Annoyingly, we can't copy values around, because they might have
   // const components (they're probably pair<const X, Y>).  We use
   // explicit destructor invocation and placement new to get around
   // this.  Arg.
- private:
   void set_value(value_type* dst, const value_type& src) {
     dst->~value_type();
     new(dst) value_type(src);
@@ -305,50 +361,74 @@ class dense_hashtable {
     assert(num_deleted == 0);
   }
 
+  bool test_deleted_key(const key_type& key) const {
+    // The num_deleted test is crucial for read(): after read(), the ht values
+    // are garbage, and we don't want to think some of them are deleted.
+    // Invariant: !use_deleted implies num_deleted is 0.
+    assert(use_deleted || num_deleted == 0);
+    return num_deleted > 0 && equals(delkey, key);
+  }
+
  public:
-  void set_deleted_key(const value_type &val) {
+  void set_deleted_key(const key_type &key) {
     // the empty indicator (if specified) and the deleted indicator
     // must be different
-    assert(!use_empty || !equals(get_key(val), get_key(emptyval)));
+    assert(!use_empty || !equals(key, get_key(emptyval)));
     // It's only safe to change what "deleted" means if we purge deleted guys
     squash_deleted();
     use_deleted = true;
-    set_value(&delval, val);
+    delkey = key;
   }
   void clear_deleted_key() {
     squash_deleted();
     use_deleted = false;
   }
+  key_type deleted_key() const {
+    assert(use_deleted);
+    return delkey;
+  }
 
   // These are public so the iterators can use them
   // True if the item at position bucknum is "deleted" marker
   bool test_deleted(size_type bucknum) const {
-    // The num_deleted test is crucial for read(): after read(), the ht values
-    // are garbage, and we don't want to think some of them are deleted.
-    return (use_deleted && num_deleted > 0 &&
-            equals(get_key(delval), get_key(table[bucknum])));
+    return test_deleted_key(get_key(table[bucknum]));
   }
   bool test_deleted(const iterator &it) const {
-    return (use_deleted && num_deleted > 0 &&
-            equals(get_key(delval), get_key(*it)));
+    return test_deleted_key(get_key(*it));
   }
   bool test_deleted(const const_iterator &it) const {
-    return (use_deleted && num_deleted > 0 &&
-            equals(get_key(delval), get_key(*it)));
+    return test_deleted_key(get_key(*it));
   }
-  // Set it so test_deleted is true.  true if object didn't used to be deleted
-  // See below (at erase()) to explain why we allow const_iterators
+
+  // Set it so test_deleted is true.  true if object didn't used to be deleted.
+  bool set_deleted(iterator &it) {
+    assert(use_deleted);             // bad if set_deleted_key() wasn't called
+    bool retval = !test_deleted(it);
+    // &* converts from iterator to value-type.
+    set_key(&(*it), delkey);
+    return retval;
+  }
+  // Set it so test_deleted is false.  true if object used to be deleted
+  bool clear_deleted(iterator &it) {
+    assert(use_deleted);             // bad if set_deleted_key() wasn't called
+    // Happens automatically when we assign something else in its place.
+    return test_deleted(it);
+  }
+
+  // We also allow to set/clear the deleted bit on a const iterator.
+  // We allow a const_iterator for the same reason you can delete a
+  // const pointer: it's convenient, and semantically you can't use
+  // 'it' after it's been deleted anyway, so its const-ness doesn't
+  // really matter.
   bool set_deleted(const_iterator &it) {
     assert(use_deleted);             // bad if set_deleted_key() wasn't called
     bool retval = !test_deleted(it);
-    // &* converts from iterator to value-type
-    set_value(const_cast<value_type*>(&(*it)), delval);
+    set_key(const_cast<value_type*>(&(*it)), delkey);
     return retval;
   }
   // Set it so test_deleted is false.  true if object used to be deleted
   bool clear_deleted(const_iterator &it) {
     assert(use_deleted);             // bad if set_deleted_key() wasn't called
-    // happens automatically when we assign something else in its place
     return test_deleted(it);
   }
 
@@ -398,7 +478,7 @@ class dense_hashtable {
     assert(!use_empty);
     // The deleted indicator (if specified) and the empty indicator
     // must be different.
-    assert(!use_deleted || !equals(get_key(val), get_key(delval)));
+    assert(!use_deleted || !equals(get_key(val), delkey));
     use_empty = true;
     set_value(&emptyval, val);
 
@@ -408,6 +488,11 @@ class dense_hashtable {
     assert(table);
     fill_range_with_empty(table, table + num_buckets);
   }
+  // TODO(sjackman): return a key_type rather than a value_type
+  value_type empty_key() const {
+    assert(use_empty);
+    return emptyval;
+  }
 
   // FUNCTIONS CONCERNING SIZE
  public:
@@ -418,6 +503,13 @@ class dense_hashtable {
   size_type bucket_count() const      { return num_buckets; }
   size_type max_bucket_count() const  { return max_size(); }
   size_type nonempty_bucket_count() const { return num_elements; }
+  // These are tr1 methods.  Their idea of 'bucket' doesn't map well to
+  // what we do.  We just say every bucket has 0 or 1 items in it.
+  size_type bucket_size(size_type i) const {
+    return begin(i) == end(i) ? 0 : 1;
+  }
+
+
 
  private:
   // Because of the above, size_type(-1) is never legal; use it for errors
@@ -428,8 +520,12 @@ class dense_hashtable {
   // If you like, you can give a min #buckets as well as a min #elts
   size_type min_size(size_type num_elts, size_type min_buckets_wanted) {
     size_type sz = HT_MIN_BUCKETS;             // min buckets allowed
-    while ( sz < min_buckets_wanted || num_elts >= sz * HT_OCCUPANCY_FLT )
+    while ( sz < min_buckets_wanted ||
+            num_elts >= static_cast<size_type>(sz * enlarge_resize_percent) ) {
+      if (sz * 2 < sz)
+        throw std::length_error("resize overflow");  // protect against overflow
       sz *= 2;
+    }
     return sz;
   }
 
@@ -439,11 +535,17 @@ class dense_hashtable {
     assert((bucket_count() & (bucket_count()-1)) == 0); // is a power of two
     assert(bucket_count() >= HT_MIN_BUCKETS);
 
-    if ( (num_elements-num_deleted) < shrink_threshold &&
-         bucket_count() > HT_MIN_BUCKETS ) {
+    // If you construct a hashtable with < HT_DEFAULT_STARTING_BUCKETS,
+    // we'll never shrink until you get relatively big, and we'll never
+    // shrink below HT_DEFAULT_STARTING_BUCKETS.  Otherwise, something
+    // like "dense_hash_set<int> x; x.insert(4); x.erase(4);" will
+    // shrink us down to HT_MIN_BUCKETS buckets, which is too small.
+    if (shrink_threshold > 0 &&
+        (num_elements-num_deleted) < shrink_threshold &&
+        bucket_count() > HT_DEFAULT_STARTING_BUCKETS ) {
       size_type sz = bucket_count() / 2;    // find how much we should shrink
-      while ( sz > HT_MIN_BUCKETS &&
-              (num_elements - num_deleted) < sz * HT_EMPTY_FLT )
+      while ( sz > HT_DEFAULT_STARTING_BUCKETS &&
+              (num_elements - num_deleted) < sz * shrink_resize_percent )
         sz /= 2;                            // stay a power of 2
       dense_hashtable tmp(*this, sz);       // Do the actual resizing
       swap(tmp);                            // now we are tmp
@@ -453,16 +555,37 @@ class dense_hashtable {
 
   // We'll let you resize a hashtable -- though this makes us copy all!
   // When you resize, you say, "make it big enough for this many more elements"
-  void resize_delta(size_type delta, size_type min_buckets_wanted = 0) {
+  void resize_delta(size_type delta) {
     if ( consider_shrink )                   // see if lots of deletes happened
       maybe_shrink();
-    if ( bucket_count() > min_buckets_wanted &&
+    if ( bucket_count() > HT_MIN_BUCKETS &&
          (num_elements + delta) <= enlarge_threshold )
       return;                                // we're ok as we are
 
-    const size_type resize_to = min_size(num_elements + delta,
-                                         min_buckets_wanted);
-    if ( resize_to > bucket_count() ) {      // we don't have enough buckets
+    // Sometimes, we need to resize just to get rid of all the
+    // "deleted" buckets that are clogging up the hashtable.  So when
+    // deciding whether to resize, count the deleted buckets (which
+    // are currently taking up room).  But later, when we decide what
+    // size to resize to, *don't* count deleted buckets, since they
+    // get discarded during the resize.
+    const size_type needed_size = min_size(num_elements + delta, 0);
+    if ( needed_size > bucket_count() ) {      // we don't have enough buckets
+      size_type resize_to = min_size(num_elements - num_deleted + delta,
+                                     bucket_count());
+      if (resize_to < needed_size) {
+        // This situation means that we have enough deleted elements,
+        // that once we purge them, we won't actually have needed to
+        // grow.  But we may want to grow anyway: if we just purge one
+        // element, say, we'll have to grow anyway next time we
+        // insert.  Might as well grow now, since we're already going
+        // through the trouble of copying (in order to purge the
+        // deleted elements).
+        if (num_elements - num_deleted + delta >=
+            static_cast<size_type>(resize_to*2 * shrink_resize_percent)) {
+          // Good, we won't be below the shrink threshhold even if we double.
+          resize_to *= 2;
+        }
+      }
       dense_hashtable tmp(*this, resize_to);
       swap(tmp);                             // now we are tmp
     }
@@ -495,7 +618,7 @@ class dense_hashtable {
   }
 
   // Used to actually do the rehashing when we grow/shrink a hashtable
-  void copy_from(const dense_hashtable &ht, size_type min_buckets_wanted = 0) {
+  void copy_from(const dense_hashtable &ht, size_type min_buckets_wanted) {
     clear();            // clear table, set num_deleted to 0
 
     // If we need to change the size of our table, do it now
@@ -527,6 +650,7 @@ class dense_hashtable {
       set_value(&table[bucknum], *it);       // copies the value to here
       num_elements++;
     }
+    num_ht_copies++;
   }
 
   // Required by the spec for hashed associative container
@@ -538,22 +662,44 @@ class dense_hashtable {
     if ( consider_shrink || req_elements == 0 )
       maybe_shrink();
     if ( req_elements > num_elements )
-      return resize_delta(req_elements - num_elements, 0);
+      return resize_delta(req_elements - num_elements);
   }
 
+  // Get and change the value of shrink_resize_percent and
+  // enlarge_resize_percent.  The description at the beginning of this
+  // file explains how to choose the values.  Setting the shrink
+  // parameter to 0.0 ensures that the table never shrinks.
+  void get_resizing_parameters(float* shrink, float* grow) const {
+    *shrink = shrink_resize_percent;
+    *grow = enlarge_resize_percent;
+  }
+  void set_resizing_parameters(float shrink, float grow) {
+    assert(shrink >= 0.0);
+    assert(grow <= 1.0);
+    if (shrink > grow/2.0f)
+      shrink = grow / 2.0f;     // otherwise we thrash hashtable size
+    shrink_resize_percent = shrink;
+    enlarge_resize_percent = grow;
+    reset_thresholds();
+  }
 
   // CONSTRUCTORS -- as required by the specs, we take a size,
   // but also let you specify a hashfunction, key comparator,
   // and key extractor.  We also define a copy constructor and =.
   // DESTRUCTOR -- needs to free the table
-  explicit dense_hashtable(size_type n = 0,
+  explicit dense_hashtable(size_type expected_max_items_in_table = 0,
                            const HashFcn& hf = HashFcn(),
                            const EqualKey& eql = EqualKey(),
-                           const ExtractKey& ext = ExtractKey())
-    : hash(hf), equals(eql), get_key(ext), num_deleted(0),
+                           const ExtractKey& ext = ExtractKey(),
+                           const SetKey& set = SetKey())
+    : hash(hf), equals(eql), get_key(ext), set_key(set), num_deleted(0),
       use_deleted(false), use_empty(false),
-      delval(), emptyval(),
-      table(NULL), num_buckets(min_size(0, n)), num_elements(0) {
+      delkey(), emptyval(), enlarge_resize_percent(HT_OCCUPANCY_FLT),
+      shrink_resize_percent(HT_EMPTY_FLT), table(NULL),
+      num_buckets(expected_max_items_in_table == 0
+                  ? HT_DEFAULT_STARTING_BUCKETS
+                  : min_size(expected_max_items_in_table, 0)),
+      num_elements(0), num_ht_copies(0) {
     // table is NULL until emptyval is set.  However, we set num_buckets
     // here so we know how much space to allocate once emptyval is set
     reset_thresholds();
@@ -561,27 +707,45 @@ class dense_hashtable {
 
   // As a convenience for resize(), we allow an optional second argument
   // which lets you make this new hashtable a different size than ht
-  dense_hashtable(const dense_hashtable& ht, size_type min_buckets_wanted = 0)
-    : hash(ht.hash), equals(ht.equals), get_key(ht.get_key), num_deleted(0),
+  dense_hashtable(const dense_hashtable& ht,
+                  size_type min_buckets_wanted = HT_DEFAULT_STARTING_BUCKETS)
+    : hash(ht.hash), equals(ht.equals),
+      get_key(ht.get_key), set_key(ht.set_key), num_deleted(0),
       use_deleted(ht.use_deleted), use_empty(ht.use_empty),
-      delval(ht.delval), emptyval(ht.emptyval),
-      table(NULL), num_buckets(0),
-      num_elements(0) {
+      delkey(ht.delkey), emptyval(ht.emptyval),
+      enlarge_resize_percent(ht.enlarge_resize_percent),
+      shrink_resize_percent(ht.shrink_resize_percent), table(NULL),
+      num_buckets(0), num_elements(0), num_ht_copies(ht.num_ht_copies) {
+    if (!ht.use_empty) {
+      // If use_empty isn't set, copy_from will crash, so we do our own copying.
+      assert(ht.empty());
+      num_buckets = min_size(ht.size(), min_buckets_wanted);
+      reset_thresholds();
+      return;
+    }
     reset_thresholds();
     copy_from(ht, min_buckets_wanted);   // copy_from() ignores deleted entries
   }
 
   dense_hashtable& operator= (const dense_hashtable& ht) {
     if (&ht == this)  return *this;        // don't copy onto ourselves
-    clear();
+    if (!ht.use_empty) {
+      assert(ht.empty());
+      dense_hashtable empty_table(ht);  // empty table with ht's thresholds
+      this->swap(empty_table);
+      return *this;
+    }
     hash = ht.hash;
     equals = ht.equals;
     get_key = ht.get_key;
+    set_key = ht.set_key;
     use_deleted = ht.use_deleted;
     use_empty = ht.use_empty;
-    set_value(&delval, ht.delval);
+    delkey = ht.delkey;
     set_value(&emptyval, ht.emptyval);
-    copy_from(ht);                         // sets num_deleted to 0 too
+    enlarge_resize_percent = ht.enlarge_resize_percent;
+    shrink_resize_percent = ht.shrink_resize_percent;
+    copy_from(ht, HT_MIN_BUCKETS);  // calls clear and sets num_deleted to 0 too
     return *this;
   }
 
@@ -597,14 +761,13 @@ class dense_hashtable {
     STL_NAMESPACE::swap(hash, ht.hash);
     STL_NAMESPACE::swap(equals, ht.equals);
     STL_NAMESPACE::swap(get_key, ht.get_key);
+    STL_NAMESPACE::swap(set_key, ht.set_key);
     STL_NAMESPACE::swap(num_deleted, ht.num_deleted);
     STL_NAMESPACE::swap(use_deleted, ht.use_deleted);
     STL_NAMESPACE::swap(use_empty, ht.use_empty);
-    { value_type tmp;     // for annoying reasons, swap() doesn't work
-      set_value(&tmp, delval);
-      set_value(&delval, ht.delval);
-      set_value(&ht.delval, tmp);
-    }
+    STL_NAMESPACE::swap(enlarge_resize_percent, ht.enlarge_resize_percent);
+    STL_NAMESPACE::swap(shrink_resize_percent, ht.shrink_resize_percent);
+    STL_NAMESPACE::swap(delkey, ht.delkey);
     { value_type tmp;     // for annoying reasons, swap() doesn't work
       set_value(&tmp, emptyval);
       set_value(&emptyval, ht.emptyval);
@@ -613,17 +776,30 @@ class dense_hashtable {
     STL_NAMESPACE::swap(table, ht.table);
     STL_NAMESPACE::swap(num_buckets, ht.num_buckets);
     STL_NAMESPACE::swap(num_elements, ht.num_elements);
+    STL_NAMESPACE::swap(num_ht_copies, ht.num_ht_copies);
     reset_thresholds();
     ht.reset_thresholds();
   }
 
   // It's always nice to be able to clear a table without deallocating it
   void clear() {
+    const size_type new_num_buckets = min_size(0,0);
+    if (num_elements == 0 &&
+        num_deleted == 0 &&
+        new_num_buckets == num_buckets) {
+      // Table is already empty, and the number of buckets is already as we
+      // desire, so nothing to do.
+      return;
+    }
     if (table)
       destroy_buckets(0, num_buckets);
-    num_buckets = min_size(0,0);          // our new size
-    reset_thresholds();
-    table = (value_type *) realloc(table, num_buckets * sizeof(*table));
+    if (!table || (new_num_buckets != num_buckets)) {
+      // Recompute the resize thresholds and realloc the table only if we're
+      // actually changing its size.
+      num_buckets = new_num_buckets;          // our new size
+      reset_thresholds();
+      table = (value_type *) realloc(table, num_buckets * sizeof(*table));
+    }
     assert(table);
     fill_range_with_empty(table, table + num_buckets);
     num_elements = 0;
@@ -694,6 +870,13 @@ class dense_hashtable {
       return const_iterator(this, table + pos.first, table+num_buckets, false);
   }
 
+  // This is a tr1 method: the bucket a given key is in, or what bucket
+  // it would be put in, if it were to be inserted.  Shrug.
+  size_type bucket(const key_type& key) const {
+    pair<size_type, size_type> pos = find_position(key);
+    return pos.first == ILLEGAL_BUCKET ? pos.second : pos.first;
+  }
+
   // Counts how many elements have key key.  For maps, it's either 0 or 1.
   size_type count(const key_type &key) const {
     pair<size_type, size_type> pos = find_position(key);
@@ -702,12 +885,22 @@ class dense_hashtable {
 
   // Likewise, equal_range doesn't really make sense for us.  Oh well.
   pair<iterator,iterator> equal_range(const key_type& key) {
-    const iterator pos = find(key);      // either an iterator or end
-    return pair<iterator,iterator>(pos, pos);
+    iterator pos = find(key);      // either an iterator or end
+    if (pos == end()) {
+      return pair<iterator,iterator>(pos, pos);
+    } else {
+      const iterator startpos = pos++;
+      return pair<iterator,iterator>(startpos, pos);
+    }
   }
   pair<const_iterator,const_iterator> equal_range(const key_type& key) const {
-    const const_iterator pos = find(key);      // either an iterator or end
-    return pair<iterator,iterator>(pos, pos);
+    const_iterator pos = find(key);      // either an iterator or end
+    if (pos == end()) {
+      return pair<const_iterator,const_iterator>(pos, pos);
+    } else {
+      const const_iterator startpos = pos++;
+      return pair<const_iterator,const_iterator>(startpos, pos);
+    }
   }
 
 
@@ -715,6 +908,9 @@ class dense_hashtable {
  private:
   // If you know *this is big enough to hold obj, use this routine
   pair<iterator, bool> insert_noresize(const value_type& obj) {
+    // First, double-check we're not inserting delkey or emptyval
+    assert(!use_empty || !equals(get_key(obj), get_key(emptyval)));
+    assert(!use_deleted || !equals(get_key(obj), delkey));
     const pair<size_type,size_type> pos = find_position(get_key(obj));
     if ( pos.first != ILLEGAL_BUCKET) {      // object was already there
       return pair<iterator,bool>(iterator(this, table + pos.first,
@@ -772,6 +968,9 @@ class dense_hashtable {
 
   // DELETION ROUTINES
   size_type erase(const key_type& key) {
+    // First, double-check we're not trying to erase delkey or emptyval
+    assert(!use_empty || !equals(key, get_key(emptyval)));
+    assert(!use_deleted || !equals(key, delkey));
     const_iterator pos = find(key);   // shrug: shouldn't need to be const
     if ( pos != end() ) {
       assert(!test_deleted(pos));  // or find() shouldn't have returned it
@@ -784,10 +983,8 @@ class dense_hashtable {
     }
   }
 
-  // This is really evil: really it should be iterator, not const_iterator.
-  // But...the only reason keys are const is to allow lookup.
-  // Since that's a moot issue for deleted keys, we allow const_iterators
-  void erase(const_iterator pos) {
+  // We return the iterator past the deleted item.
+  void erase(iterator pos) {
     if ( pos == end() ) return;    // sanity check
     if ( set_deleted(pos) ) {      // true if object has been newly deleted
       ++num_deleted;
@@ -795,6 +992,26 @@ class dense_hashtable {
     }
   }
 
+  void erase(iterator f, iterator l) {
+    for ( ; f != l; ++f) {
+      if ( set_deleted(f)  )       // should always be true
+        ++num_deleted;
+    }
+    consider_shrink = true;        // will think about shrink after next insert
+  }
+
+  // We allow you to erase a const_iterator just like we allow you to
+  // erase an iterator.  This is in parallel to 'delete': you can delete
+  // a const pointer just like a non-const pointer.  The logic is that
+  // you can't use the object after it's erased anyway, so it doesn't matter
+  // if it's const or not.
+  void erase(const_iterator pos) {
+    if ( pos == end() ) return;    // sanity check
+    if ( set_deleted(pos) ) {      // true if object has been newly deleted
+      ++num_deleted;
+      consider_shrink = true;      // will think about shrink after next insert
+    }
+  }
   void erase(const_iterator f, const_iterator l) {
     for ( ; f != l; ++f) {
       if ( set_deleted(f)  )       // should always be true
@@ -834,7 +1051,7 @@ class dense_hashtable {
   //
   // NOTE: These functions are currently TODO.  They've not been implemented.
   bool write_metadata(FILE *fp) {
-    squash_deleted();           // so we don't have to worry about delval
+    squash_deleted();           // so we don't have to worry about delkey
     return false;               // TODO
   }
 
@@ -883,48 +1100,58 @@ class dense_hashtable {
   hasher hash;                      // required by hashed_associative_container
   key_equal equals;
   ExtractKey get_key;
+  SetKey set_key;
   size_type num_deleted;        // how many occupied buckets are marked deleted
-  bool use_deleted;                          // false until delval has been set
+  bool use_deleted;                          // false until delkey has been set
   bool use_empty;                          // you must do this before you start
-  value_type delval;                         // which key marks deleted entries
+  // TODO(csilvers): make a pointer, and get rid of use_deleted (benchmark!)
+  key_type delkey;                           // which key marks deleted entries
   value_type emptyval;                        // which key marks unused entries
+  float enlarge_resize_percent;                       // how full before resize
+  float shrink_resize_percent;                       // how empty before resize
+  size_type shrink_threshold;            // num_buckets * shrink_resize_percent
+  size_type enlarge_threshold;          // num_buckets * enlarge_resize_percent
   value_type *table;
   size_type num_buckets;
   size_type num_elements;
-  size_type shrink_threshold;                     // num_buckets * HT_EMPTY_FLT
-  size_type enlarge_threshold;                // num_buckets * HT_OCCUPANCY_FLT
+  int num_ht_copies;             // a statistics counter incremented every Copy
   bool consider_shrink;   // true if we should try to shrink before next insert
 
   void reset_thresholds() {
-    enlarge_threshold = static_cast<size_type>(num_buckets*HT_OCCUPANCY_FLT);
-    shrink_threshold = static_cast<size_type>(num_buckets*HT_EMPTY_FLT);
+    enlarge_threshold = static_cast<size_type>(num_buckets
+                                               * enlarge_resize_percent);
+    shrink_threshold = static_cast<size_type>(num_buckets
+                                              * shrink_resize_percent);
     consider_shrink = false;   // whatever caused us to reset already considered
   }
 };
 
 // We need a global swap as well
-template <class V, class K, class HF, class ExK, class EqK, class A>
-inline void swap(dense_hashtable<V,K,HF,ExK,EqK,A> &x,
-                 dense_hashtable<V,K,HF,ExK,EqK,A> &y) {
+template <class V, class K, class HF, class ExK, class SetK, class EqK, class A>
+inline void swap(dense_hashtable<V,K,HF,ExK,SetK,EqK,A> &x,
+                 dense_hashtable<V,K,HF,ExK,SetK,EqK,A> &y) {
   x.swap(y);
 }
 
 #undef JUMP_
 
-template <class V, class K, class HF, class ExK, class EqK, class A>
-const typename dense_hashtable<V,K,HF,ExK,EqK,A>::size_type
-  dense_hashtable<V,K,HF,ExK,EqK,A>::ILLEGAL_BUCKET;
+template <class V, class K, class HF, class ExK, class SetK, class EqK, class A>
+const typename dense_hashtable<V,K,HF,ExK,SetK,EqK,A>::size_type
+dense_hashtable<V,K,HF,ExK,SetK,EqK,A>::ILLEGAL_BUCKET;
 
 // How full we let the table get before we resize.  Knuth says .8 is
-// good -- higher causes us to probe too much, though saves memory
-template <class V, class K, class HF, class ExK, class EqK, class A>
-const float dense_hashtable<V,K,HF,ExK,EqK,A>::HT_OCCUPANCY_FLT = 0.5f;
+// good -- higher causes us to probe too much, though saves memory.
+// However, we go with .5, getting better performance at the cost of
+// more space (a trade-off densehashtable explicitly chooses to make).
+// Feel free to play around with different values, though.
+template <class V, class K, class HF, class ExK, class SetK, class EqK, class A>
+const float dense_hashtable<V,K,HF,ExK,SetK,EqK,A>::HT_OCCUPANCY_FLT = 0.5f;
 
 // How empty we let the table get before we resize lower.
 // It should be less than OCCUPANCY_FLT / 2 or we thrash resizing
-template <class V, class K, class HF, class ExK, class EqK, class A>
-const float dense_hashtable<V,K,HF,ExK,EqK,A>::HT_EMPTY_FLT = 0.4f *
-dense_hashtable<V,K,HF,ExK,EqK,A>::HT_OCCUPANCY_FLT;
+template <class V, class K, class HF, class ExK, class SetK, class EqK, class A>
+const float dense_hashtable<V,K,HF,ExK,SetK,EqK,A>::HT_EMPTY_FLT
+    = 0.4f * dense_hashtable<V,K,HF,ExK,SetK,EqK,A>::HT_OCCUPANCY_FLT;
 
 _END_GOOGLE_NAMESPACE_
 
