@@ -57,17 +57,22 @@
 //
 // You probably shouldn't use this code directly.  Use
 // <google/sparse_hash_table> or <google/sparse_hash_set> instead.
-
-// You can change the following below:
-// HT_OCCUPANCY_FLT      -- how full before we double size
-// HT_EMPTY_FLT          -- how empty before  we halve size
-// HT_MIN_BUCKETS        -- smallest bucket size
+//
+// You can modify the following, below:
+// HT_OCCUPANCY_FLT            -- how full before we double size
+// HT_EMPTY_FLT                -- how empty before we halve size
+// HT_MIN_BUCKETS              -- smallest bucket size
+// HT_DEFAULT_STARTING_BUCKETS -- default bucket size at construct-time
+//
+// You can also change enlarge_resize_percent (which defaults to
+// HT_OCCUPANCY_FLT), and shrink_resize_percent (which defaults to
+// HT_EMPTY_FLT) with set_resizing_parameters().
 //
 // How to decide what values to use?
-// HT_EMPTY_FLT's default of .4 * OCCUPANCY_FLT, is probably good.
+// shrink_resize_percent's default of .4 * OCCUPANCY_FLT, is probably good.
 // HT_MIN_BUCKETS is probably unnecessary since you can specify
 // (indirectly) the starting number of buckets at construct-time.
-// For HT_OCCUPANCY_FLT, you can use this chart to try to trade-off
+// For enlarge_resize_percent, you can use this chart to try to trade-off
 // expected lookup time to the space taken up.  By default, this
 // code uses quadratic probing, though you can change it to linear
 // via _JUMP below if you really want to.
@@ -77,7 +82,7 @@
 // Quadratic collision resolution   1 - ln(1-L) - L/2    1/(1-L) - L - ln(1-L)
 // Linear collision resolution     [1+1/(1-L)]/2         [1+1/(1-L)2]/2
 //
-// -- HT_OCCUPANCY_FLT --         0.10  0.50  0.60  0.75  0.80  0.90  0.99
+// -- enlarge_resize_percent --         0.10  0.50  0.60  0.75  0.80  0.90  0.99
 // QUADRATIC COLLISION RES.
 //    probes/successful lookup    1.05  1.44  1.62  2.01  2.21  2.85  5.11
 //    probes/unsuccessful lookup  1.11  2.19  2.82  4.64  5.81  11.4  103.6
@@ -102,12 +107,10 @@
 #define JUMP_(key, num_probes)    ( num_probes )
 
 
-// Hashtable class, used to implement the hashed associative containers
-// hash_set and hash_map.
-
 #include <google/sparsehash/sparseconfig.h>
 #include <assert.h>
 #include <algorithm>              // For swap(), eg
+#include <stdexcept>              // For length_error
 #include <iterator>               // for facts about iterator tags
 #include <utility>                // for pair<>
 #include <google/sparsetable>     // Since that's basically what we are
@@ -116,27 +119,38 @@ _START_GOOGLE_NAMESPACE_
 
 using STL_NAMESPACE::pair;
 
-// Alloc is completely ignored.  It is present as a template parameter only
-// for the sake of being compatible with the old SGI hashtable interface.
-// TODO(csilvers): is that the right thing to do?
+// Hashtable class, used to implement the hashed associative containers
+// hash_set and hash_map.
+//
+// Value: what is stored in the table (each bucket is a Value).
+// Key: something in a 1-to-1 correspondence to a Value, that can be used
+//      to search for a Value in the table (find() takes a Key).
+// HashFcn: Takes a Key and returns an integer, the more unique the better.
+// ExtractKey: given a Value, returns the unique Key associated with it.
+// SetKey: given a Value* and a Key, modifies the value such that
+//         ExtractKey(value) == key.  We guarantee this is only called
+//         with key == deleted_key.
+// EqualKey: Given two Keys, says whether they are the same (that is,
+//           if they are both associated with the same Value).
+// Alloc: STL allocator to use to allocate memory.  Currently ignored.
 
 template <class Value, class Key, class HashFcn,
-          class ExtractKey, class EqualKey, class Alloc>
+          class ExtractKey, class SetKey, class EqualKey, class Alloc>
 class sparse_hashtable;
 
-template <class V, class K, class HF, class ExK, class EqK, class A>
+template <class V, class K, class HF, class ExK, class SetK, class EqK, class A>
 struct sparse_hashtable_iterator;
 
-template <class V, class K, class HF, class ExK, class EqK, class A>
+template <class V, class K, class HF, class ExK, class SetK, class EqK, class A>
 struct sparse_hashtable_const_iterator;
 
 // As far as iterating, we're basically just a sparsetable
 // that skips over deleted elements.
-template <class V, class K, class HF, class ExK, class EqK, class A>
+template <class V, class K, class HF, class ExK, class SetK, class EqK, class A>
 struct sparse_hashtable_iterator {
  public:
-  typedef sparse_hashtable_iterator<V,K,HF,ExK,EqK,A>       iterator;
-  typedef sparse_hashtable_const_iterator<V,K,HF,ExK,EqK,A> const_iterator;
+  typedef sparse_hashtable_iterator<V,K,HF,ExK,SetK,EqK,A>       iterator;
+  typedef sparse_hashtable_const_iterator<V,K,HF,ExK,SetK,EqK,A> const_iterator;
   typedef typename sparsetable<V>::nonempty_iterator        st_iterator;
 
   typedef STL_NAMESPACE::forward_iterator_tag iterator_category;
@@ -147,7 +161,7 @@ struct sparse_hashtable_iterator {
   typedef V* pointer;
 
   // "Real" constructor and default constructor
-  sparse_hashtable_iterator(const sparse_hashtable<V,K,HF,ExK,EqK,A> *h,
+  sparse_hashtable_iterator(const sparse_hashtable<V,K,HF,ExK,SetK,EqK,A> *h,
                             st_iterator it, st_iterator it_end)
     : ht(h), pos(it), end(it_end)   { advance_past_deleted(); }
   sparse_hashtable_iterator() { }      // not ever used internally
@@ -175,16 +189,16 @@ struct sparse_hashtable_iterator {
 
 
   // The actual data
-  const sparse_hashtable<V,K,HF,ExK,EqK,A> *ht;
+  const sparse_hashtable<V,K,HF,ExK,SetK,EqK,A> *ht;
   st_iterator pos, end;
 };
 
 // Now do it all again, but with const-ness!
-template <class V, class K, class HF, class ExK, class EqK, class A>
+template <class V, class K, class HF, class ExK, class SetK, class EqK, class A>
 struct sparse_hashtable_const_iterator {
  public:
-  typedef sparse_hashtable_iterator<V,K,HF,ExK,EqK,A>       iterator;
-  typedef sparse_hashtable_const_iterator<V,K,HF,ExK,EqK,A> const_iterator;
+  typedef sparse_hashtable_iterator<V,K,HF,ExK,SetK,EqK,A>       iterator;
+  typedef sparse_hashtable_const_iterator<V,K,HF,ExK,SetK,EqK,A> const_iterator;
   typedef typename sparsetable<V>::const_nonempty_iterator  st_iterator;
 
   typedef STL_NAMESPACE::forward_iterator_tag iterator_category;
@@ -195,7 +209,7 @@ struct sparse_hashtable_const_iterator {
   typedef const V* pointer;
 
   // "Real" constructor and default constructor
-  sparse_hashtable_const_iterator(const sparse_hashtable<V,K,HF,ExK,EqK,A> *h,
+  sparse_hashtable_const_iterator(const sparse_hashtable<V,K,HF,ExK,SetK,EqK,A> *h,
                                   st_iterator it, st_iterator it_end)
     : ht(h), pos(it), end(it_end)   { advance_past_deleted(); }
   // This lets us convert regular iterators to const iterators
@@ -226,15 +240,15 @@ struct sparse_hashtable_const_iterator {
 
 
   // The actual data
-  const sparse_hashtable<V,K,HF,ExK,EqK,A> *ht;
+  const sparse_hashtable<V,K,HF,ExK,SetK,EqK,A> *ht;
   st_iterator pos, end;
 };
 
 // And once again, but this time freeing up memory as we iterate
-template <class V, class K, class HF, class ExK, class EqK, class A>
+template <class V, class K, class HF, class ExK, class SetK, class EqK, class A>
 struct sparse_hashtable_destructive_iterator {
  public:
-  typedef sparse_hashtable_destructive_iterator<V,K,HF,ExK,EqK,A> iterator;
+  typedef sparse_hashtable_destructive_iterator<V,K,HF,ExK,SetK,EqK,A> iterator;
   typedef typename sparsetable<V>::destructive_iterator     st_iterator;
 
   typedef STL_NAMESPACE::forward_iterator_tag iterator_category;
@@ -246,7 +260,7 @@ struct sparse_hashtable_destructive_iterator {
 
   // "Real" constructor and default constructor
   sparse_hashtable_destructive_iterator(const
-                                        sparse_hashtable<V,K,HF,ExK,EqK,A> *h,
+                                        sparse_hashtable<V,K,HF,ExK,SetK,EqK,A> *h,
                                         st_iterator it, st_iterator it_end)
     : ht(h), pos(it), end(it_end)   { advance_past_deleted(); }
   sparse_hashtable_destructive_iterator() { }          // never used internally
@@ -274,13 +288,13 @@ struct sparse_hashtable_destructive_iterator {
 
 
   // The actual data
-  const sparse_hashtable<V,K,HF,ExK,EqK,A> *ht;
+  const sparse_hashtable<V,K,HF,ExK,SetK,EqK,A> *ht;
   st_iterator pos, end;
 };
 
 
 template <class Value, class Key, class HashFcn,
-          class ExtractKey, class EqualKey, class Alloc>
+          class ExtractKey, class SetKey, class EqualKey, class Alloc>
 class sparse_hashtable {
  public:
   typedef Key key_type;
@@ -294,23 +308,28 @@ class sparse_hashtable {
   typedef const value_type* const_pointer;
   typedef value_type&       reference;
   typedef const value_type& const_reference;
-  typedef sparse_hashtable_iterator<Value, Key, HashFcn,
-                                    ExtractKey, EqualKey, Alloc>
+  typedef sparse_hashtable_iterator<Value, Key, HashFcn, ExtractKey,
+                                    SetKey, EqualKey, Alloc>
   iterator;
 
-  typedef sparse_hashtable_const_iterator<Value, Key, HashFcn,
-                                          ExtractKey, EqualKey, Alloc>
+  typedef sparse_hashtable_const_iterator<Value, Key, HashFcn, ExtractKey,
+                                          SetKey, EqualKey, Alloc>
   const_iterator;
 
-  typedef sparse_hashtable_destructive_iterator<Value, Key, HashFcn,
-                                                ExtractKey, EqualKey, Alloc>
+  typedef sparse_hashtable_destructive_iterator<Value, Key, HashFcn, ExtractKey,
+                                                SetKey, EqualKey, Alloc>
   destructive_iterator;
 
-  // How full we let the table get before we resize.  Knuth says .8 is
-  // good -- higher causes us to probe too much, though saves memory
+  // These come from tr1.  For us they're the same as regular iterators.
+  typedef iterator local_iterator;
+  typedef const_iterator const_local_iterator;
+
+  // How full we let the table get before we resize, by default.
+  // Knuth says .8 is good -- higher causes us to probe too much,
+  // though it saves memory.
   static const float HT_OCCUPANCY_FLT; // = 0.8f;
 
-  // How empty we let the table get before we resize lower.
+  // How empty we let the table get before we resize lower, by default.
   // It should be less than OCCUPANCY_FLT / 2 or we thrash resizing
   static const float HT_EMPTY_FLT; // = 0.4 * HT_OCCUPANCY_FLT;
 
@@ -318,8 +337,12 @@ class sparse_hashtable {
   // Must be a power of two, and at least 4.
   // Note, however, that for a given hashtable, the minimum size is
   // determined by the first constructor arg, and may be >HT_MIN_BUCKETS.
-  static const size_t HT_MIN_BUCKETS = 32;
+  static const size_t HT_MIN_BUCKETS = 4;
 
+  // By default, if you don't specify a hashtable size at
+  // construction-time, we use this size.  Must be a power of two, and
+  // at least HT_MIN_BUCKETS.
+  static const size_t HT_DEFAULT_STARTING_BUCKETS = 32;
 
   // ITERATOR FUNCTIONS
   iterator begin()             { return iterator(this, table.nonempty_begin(),
@@ -333,6 +356,38 @@ class sparse_hashtable {
                                                        table.nonempty_end(),
                                                        table.nonempty_end()); }
 
+  // These come from tr1 unordered_map.  They iterate over 'bucket' n.
+  // For sparsehashtable, we could consider each 'group' to be a bucket,
+  // I guess, but I don't really see the point.  We'll just consider
+  // bucket n to be the n-th element of the sparsetable, if it's occupied,
+  // or some empty element, otherwise.
+  local_iterator begin(size_type i) {
+    if (table.test(i))
+      return local_iterator(this, table.get_iter(i), table.nonempty_end());
+    else
+      return local_iterator(this, table.nonempty_end(), table.nonempty_end());
+  }
+  local_iterator end(size_type i) {
+    local_iterator it = begin(i);
+    if (table.test(i) && !test_deleted(i))
+      ++it;
+    return it;
+  }
+  const_local_iterator begin(size_type i) const {
+    if (table.test(i))
+      return const_local_iterator(this, table.get_iter(i),
+                                  table.nonempty_end());
+    else
+      return const_local_iterator(this, table.nonempty_end(),
+                                  table.nonempty_end());
+  }
+  const_local_iterator end(size_type i) const {
+    const_local_iterator it = begin(i);
+    if (table.test(i) && !test_deleted(i))
+      ++it;
+    return it;
+  }
+
   // This is used when resizing
   destructive_iterator destructive_begin() {
     return destructive_iterator(this, table.destructive_begin(),
@@ -348,12 +403,15 @@ class sparse_hashtable {
   hasher hash_funct() const { return hash; }
   key_equal key_eq() const  { return equals; }
 
+  // Accessor function for statistics gathering.
+  int num_table_copies() const { return num_ht_copies; }
+
+ private:
   // We need to copy values when we set the special marker for deleted
   // elements, but, annoyingly, we can't just use the copy assignment
   // operator because value_type might not be assignable (it's often
   // pair<const X, Y>).  We use explicit destructor invocation and
   // placement new to get around this.  Arg.
- private:
   void set_value(value_type* dst, const value_type src) {
     dst->~value_type();   // delete the old value, if any
     new(dst) value_type(src);
@@ -382,16 +440,20 @@ class sparse_hashtable {
   }
 
  public:
-  void set_deleted_key(const value_type &val) {
+  void set_deleted_key(const key_type &key) {
     // It's only safe to change what "deleted" means if we purge deleted guys
     squash_deleted();
     use_deleted = true;
-    set_value(&delval, val);        // save the key (and rest of val too)
+    delkey = key;
   }
   void clear_deleted_key() {
     squash_deleted();
     use_deleted = false;
   }
+  key_type deleted_key() const {
+    assert(use_deleted);
+    return delkey;
+  }
 
   // These are public so the iterators can use them
   // True if the item at position bucknum is "deleted" marker
@@ -399,33 +461,48 @@ class sparse_hashtable {
     // The num_deleted test is crucial for read(): after read(), the ht values
     // are garbage, and we don't want to think some of them are deleted.
     return (use_deleted && num_deleted > 0 && table.test(bucknum) &&
-            equals(get_key(delval), get_key(table.get(bucknum))));
+            equals(delkey, get_key(table.unsafe_get(bucknum))));
   }
   bool test_deleted(const iterator &it) const {
     return (use_deleted && num_deleted > 0 &&
-            equals(get_key(delval), get_key(*it)));
+            equals(delkey, get_key(*it)));
   }
   bool test_deleted(const const_iterator &it) const {
     return (use_deleted && num_deleted > 0 &&
-            equals(get_key(delval), get_key(*it)));
+            equals(delkey, get_key(*it)));
   }
   bool test_deleted(const destructive_iterator &it) const {
     return (use_deleted && num_deleted > 0 &&
-            equals(get_key(delval), get_key(*it)));
+            equals(delkey, get_key(*it)));
   }
-  // Set it so test_deleted is true.  true if object didn't used to be deleted
-  // See below (at erase()) to explain why we allow const_iterators
+  // Set it so test_deleted is true.  true if object didn't used to be deleted.
+  bool set_deleted(iterator &it) {
+    assert(use_deleted);
+    bool retval = !test_deleted(it);
+    // &* converts from iterator to value-type.
+    set_key(&(*it), delkey);
+    return retval;
+  }
+  // Set it so test_deleted is false.  true if object used to be deleted.
+  bool clear_deleted(iterator &it) {
+    assert(use_deleted);
+    // Happens automatically when we assign something else in its place.
+    return test_deleted(it);
+  }
+
+  // We also allow to set/clear the deleted bit on a const iterator.
+  // We allow a const_iterator for the same reason you can delete a
+  // const pointer: it's convenient, and semantically you can't use
+  // 'it' after it's been deleted anyway, so its const-ness doesn't
+  // really matter.
   bool set_deleted(const_iterator &it) {
     assert(use_deleted);             // bad if set_deleted_key() wasn't called
     bool retval = !test_deleted(it);
-    // &* converts from iterator to value-type
-    set_value(const_cast<value_type*>(&(*it)), delval);
+    set_key(const_cast<value_type*>(&(*it)), delkey);
     return retval;
   }
-  // Set it so test_deleted is false.  true if object used to be deleted
   bool clear_deleted(const_iterator &it) {
     assert(use_deleted);             // bad if set_deleted_key() wasn't called
-    // happens automatically when we assign something else in its place
     return test_deleted(it);
   }
 
@@ -433,10 +510,16 @@ class sparse_hashtable {
   // FUNCTIONS CONCERNING SIZE
   size_type size() const      { return table.num_nonempty() - num_deleted; }
   // Buckets are always a power of 2
-  size_type max_size() const  { return (size_type(-1) >> 1U) + 1; }
-  bool empty() const          { return size() == 0; }
+  size_type max_size() const          { return (size_type(-1) >> 1U) + 1; }
+  bool empty() const                  { return size() == 0; }
   size_type bucket_count() const      { return table.size(); }
   size_type max_bucket_count() const  { return max_size(); }
+  // These are tr1 methods.  Their idea of 'bucket' doesn't map well to
+  // what we do.  We just say every bucket has 0 or 1 items in it.
+  size_type bucket_size(size_type i) const {
+    return begin(i) == end(i) ? 0 : 1;
+  }
+
 
  private:
   // Because of the above, size_type(-1) is never legal; use it for errors
@@ -446,9 +529,13 @@ class sparse_hashtable {
   // This is the smallest size a hashtable can be without being too crowded
   // If you like, you can give a min #buckets as well as a min #elts
   size_type min_size(size_type num_elts, size_type min_buckets_wanted) {
-    size_type sz = HT_MIN_BUCKETS;
-    while ( sz < min_buckets_wanted || num_elts >= sz * HT_OCCUPANCY_FLT )
+    size_type sz = HT_MIN_BUCKETS;             // min buckets allowed
+    while ( sz < min_buckets_wanted ||
+            num_elts >= static_cast<size_type>(sz * enlarge_resize_percent) ) {
+      if (sz * 2 < sz)
+        throw std::length_error("resize overflow");  // protect against overflow
       sz *= 2;
+    }
     return sz;
   }
 
@@ -458,11 +545,18 @@ class sparse_hashtable {
     assert((bucket_count() & (bucket_count()-1)) == 0); // is a power of two
     assert(bucket_count() >= HT_MIN_BUCKETS);
 
-    if ( (table.num_nonempty()-num_deleted) <= shrink_threshold &&
-         bucket_count() > HT_MIN_BUCKETS ) {
+    // If you construct a hashtable with < HT_DEFAULT_STARTING_BUCKETS,
+    // we'll never shrink until you get relatively big, and we'll never
+    // shrink below HT_DEFAULT_STARTING_BUCKETS.  Otherwise, something
+    // like "dense_hash_set<int> x; x.insert(4); x.erase(4);" will
+    // shrink us down to HT_MIN_BUCKETS buckets, which is too small.
+    if (shrink_threshold > 0 &&
+        (table.num_nonempty()-num_deleted) < shrink_threshold &&
+        bucket_count() > HT_DEFAULT_STARTING_BUCKETS) {
       size_type sz = bucket_count() / 2;    // find how much we should shrink
-      while ( sz > HT_MIN_BUCKETS &&
-              (table.num_nonempty() - num_deleted) <= sz * HT_EMPTY_FLT )
+      while ( sz > HT_DEFAULT_STARTING_BUCKETS &&
+              (table.num_nonempty() - num_deleted) <= sz *
+              shrink_resize_percent )
         sz /= 2;                            // stay a power of 2
       sparse_hashtable tmp(MoveDontCopy, *this, sz);
       swap(tmp);                            // now we are tmp
@@ -472,23 +566,45 @@ class sparse_hashtable {
 
   // We'll let you resize a hashtable -- though this makes us copy all!
   // When you resize, you say, "make it big enough for this many more elements"
-  void resize_delta(size_type delta, size_type min_buckets_wanted = 0) {
+  void resize_delta(size_type delta) {
     if ( consider_shrink )                   // see if lots of deletes happened
       maybe_shrink();
-    if ( bucket_count() > min_buckets_wanted &&
+    if ( bucket_count() >= HT_MIN_BUCKETS &&
          (table.num_nonempty() + delta) <= enlarge_threshold )
       return;                                // we're ok as we are
 
-    const size_type resize_to = min_size(table.num_nonempty() + delta,
-                                         min_buckets_wanted);
-    if ( resize_to > bucket_count() ) {      // we don't have enough buckets
+    // Sometimes, we need to resize just to get rid of all the
+    // "deleted" buckets that are clogging up the hashtable.  So when
+    // deciding whether to resize, count the deleted buckets (which
+    // are currently taking up room).  But later, when we decide what
+    // size to resize to, *don't* count deleted buckets, since they
+    // get discarded during the resize.
+    const size_type needed_size = min_size(table.num_nonempty() + delta, 0);
+    if ( needed_size > bucket_count() ) {      // we don't have enough buckets
+      size_type resize_to = min_size(table.num_nonempty() - num_deleted + delta,
+                                     bucket_count());
+      if (resize_to < needed_size) {
+        // This situation means that we have enough deleted elements,
+        // that once we purge them, we won't actually have needed to
+        // grow.  But we may want to grow anyway: if we just purge one
+        // element, say, we'll have to grow anyway next time we
+        // insert.  Might as well grow now, since we're already going
+        // through the trouble of copying (in order to purge the
+        // deleted elements).
+        if (table.num_nonempty() - num_deleted + delta >=
+            static_cast<size_type>(resize_to*2 * shrink_resize_percent)) {
+          // Good, we won't be below the shrink threshhold even if we double.
+          resize_to *= 2;
+        }
+      }
+
       sparse_hashtable tmp(MoveDontCopy, *this, resize_to);
       swap(tmp);                             // now we are tmp
     }
   }
 
   // Used to actually do the rehashing when we grow/shrink a hashtable
-  void copy_from(const sparse_hashtable &ht, size_type min_buckets_wanted = 0) {
+  void copy_from(const sparse_hashtable &ht, size_type min_buckets_wanted) {
     clear();            // clear table, set num_deleted to 0
 
     // If we need to change the size of our table, do it now
@@ -514,13 +630,14 @@ class sparse_hashtable {
       }
       table.set(bucknum, *it);               // copies the value to here
     }
+    num_ht_copies++;
   }
 
   // Implementation is like copy_from, but it destroys the table of the
   // "from" guy by freeing sparsetable memory as we iterate.  This is
   // useful in resizing, since we're throwing away the "from" guy anyway.
   void move_from(MoveDontCopyT mover, sparse_hashtable &ht,
-                 size_type min_buckets_wanted = 0) {
+                 size_type min_buckets_wanted) {
     clear();            // clear table, set num_deleted to 0
 
     // If we need to change the size of our table, do it now
@@ -551,6 +668,7 @@ class sparse_hashtable {
       }
       table.set(bucknum, *it);               // copies the value to here
     }
+    num_ht_copies++;
   }
 
 
@@ -563,20 +681,43 @@ class sparse_hashtable {
     if ( consider_shrink || req_elements == 0 )
       maybe_shrink();
     if ( req_elements > table.num_nonempty() )    // we only grow
-      resize_delta(req_elements - table.num_nonempty(), 0);
+      resize_delta(req_elements - table.num_nonempty());
   }
 
+  // Get and change the value of shrink_resize_percent and
+  // enlarge_resize_percent.  The description at the beginning of this
+  // file explains how to choose the values.  Setting the shrink
+  // parameter to 0.0 ensures that the table never shrinks.
+  void get_resizing_parameters(float* shrink, float* grow) const {
+    *shrink = shrink_resize_percent;
+    *grow = enlarge_resize_percent;
+  }
+  void set_resizing_parameters(float shrink, float grow) {
+    assert(shrink >= 0.0);
+    assert(grow <= 1.0);
+    if (shrink > grow/2.0f)
+      shrink = grow / 2.0f;     // otherwise we thrash hashtable size
+    shrink_resize_percent = shrink;
+    enlarge_resize_percent = grow;
+    reset_thresholds();
+  }
 
   // CONSTRUCTORS -- as required by the specs, we take a size,
   // but also let you specify a hashfunction, key comparator,
   // and key extractor.  We also define a copy constructor and =.
   // DESTRUCTOR -- the default is fine, surprisingly.
-  explicit sparse_hashtable(size_type n = 0,
+  explicit sparse_hashtable(size_type expected_max_items_in_table = 0,
                             const HashFcn& hf = HashFcn(),
                             const EqualKey& eql = EqualKey(),
+                            const SetKey& set = SetKey(),
                             const ExtractKey& ext = ExtractKey())
-    : hash(hf), equals(eql), get_key(ext), num_deleted(0),
-      use_deleted(false), delval(), table(min_size(0, n)) {    // start small
+    : hash(hf), equals(eql), get_key(ext), set_key(set), num_deleted(0),
+      use_deleted(false), delkey(), enlarge_resize_percent(HT_OCCUPANCY_FLT),
+      shrink_resize_percent(HT_EMPTY_FLT),
+      table(expected_max_items_in_table == 0
+            ? HT_DEFAULT_STARTING_BUCKETS
+            : min_size(expected_max_items_in_table, 0)),
+      num_ht_copies(0) {
     reset_thresholds();
   }
 
@@ -584,29 +725,37 @@ class sparse_hashtable {
   // which lets you make this new hashtable a different size than ht.
   // We also provide a mechanism of saying you want to "move" the ht argument
   // into us instead of copying.
-  sparse_hashtable(const sparse_hashtable& ht, size_type min_buckets_wanted = 0)
-    : hash(ht.hash), equals(ht.equals), get_key(ht.get_key),
-      num_deleted(0), use_deleted(ht.use_deleted), delval(ht.delval), table() {
+  sparse_hashtable(const sparse_hashtable& ht,
+                   size_type min_buckets_wanted = HT_DEFAULT_STARTING_BUCKETS)
+    : hash(ht.hash), equals(ht.equals),
+      get_key(ht.get_key), set_key(ht.set_key), num_deleted(0),
+      use_deleted(ht.use_deleted), delkey(ht.delkey),
+      enlarge_resize_percent(ht.enlarge_resize_percent),
+      shrink_resize_percent(ht.shrink_resize_percent),
+      table(), num_ht_copies(ht.num_ht_copies) {
     reset_thresholds();
     copy_from(ht, min_buckets_wanted);   // copy_from() ignores deleted entries
   }
   sparse_hashtable(MoveDontCopyT mover, sparse_hashtable& ht,
-                   size_type min_buckets_wanted=0)
+                   size_type min_buckets_wanted = HT_DEFAULT_STARTING_BUCKETS)
     : hash(ht.hash), equals(ht.equals), get_key(ht.get_key),
-      num_deleted(0), use_deleted(ht.use_deleted), delval(ht.delval), table() {
+      num_deleted(0), use_deleted(ht.use_deleted), delkey(ht.delkey),
+      enlarge_resize_percent(ht.enlarge_resize_percent),
+      shrink_resize_percent(ht.shrink_resize_percent),
+      table(), num_ht_copies(ht.num_ht_copies) {
     reset_thresholds();
     move_from(mover, ht, min_buckets_wanted);  // ignores deleted entries
   }
 
   sparse_hashtable& operator= (const sparse_hashtable& ht) {
     if (&ht == this)  return *this;        // don't copy onto ourselves
-    clear();
     hash = ht.hash;
     equals = ht.equals;
     get_key = ht.get_key;
+    set_key = ht.set_key;
     use_deleted = ht.use_deleted;
-    set_value(&delval, ht.delval);
-    copy_from(ht);                         // sets num_deleted to 0 too
+    delkey = ht.delkey;
+    copy_from(ht, HT_MIN_BUCKETS);  // calls clear and sets num_deleted to 0 too
     return *this;
   }
 
@@ -615,21 +764,23 @@ class sparse_hashtable {
     STL_NAMESPACE::swap(hash, ht.hash);
     STL_NAMESPACE::swap(equals, ht.equals);
     STL_NAMESPACE::swap(get_key, ht.get_key);
+    STL_NAMESPACE::swap(set_key, ht.set_key);
     STL_NAMESPACE::swap(num_deleted, ht.num_deleted);
     STL_NAMESPACE::swap(use_deleted, ht.use_deleted);
-    { value_type tmp;     // for annoying reasons, swap() doesn't work
-      set_value(&tmp, delval);
-      set_value(&delval, ht.delval);
-      set_value(&ht.delval, tmp);
-    }
+    STL_NAMESPACE::swap(enlarge_resize_percent, ht.enlarge_resize_percent);
+    STL_NAMESPACE::swap(shrink_resize_percent, ht.shrink_resize_percent);
+    STL_NAMESPACE::swap(delkey, ht.delkey);
     table.swap(ht.table);
+    STL_NAMESPACE::swap(num_ht_copies, ht.num_ht_copies);
     reset_thresholds();
     ht.reset_thresholds();
   }
 
   // It's always nice to be able to clear a table without deallocating it
   void clear() {
-    table.clear();
+    if (!empty() || (num_deleted != 0)) {
+      table.clear();
+    }
     reset_thresholds();
     num_deleted = 0;
   }
@@ -660,7 +811,7 @@ class sparse_hashtable {
         if ( insert_pos == ILLEGAL_BUCKET )
           insert_pos = bucknum;
 
-      } else if ( equals(key, get_key(table.get(bucknum))) ) {
+      } else if ( equals(key, get_key(table.unsafe_get(bucknum))) ) {
         SPARSEHASH_STAT_UPDATE(total_probes += num_probes);
         return pair<size_type,size_type>(bucknum, ILLEGAL_BUCKET);
       }
@@ -690,6 +841,13 @@ class sparse_hashtable {
                             table.get_iter(pos.first), table.nonempty_end());
   }
 
+  // This is a tr1 method: the bucket a given key is in, or what bucket
+  // it would be put in, if it were to be inserted.  Shrug.
+  size_type bucket(const key_type& key) const {
+    pair<size_type, size_type> pos = find_position(key);
+    return pos.first == ILLEGAL_BUCKET ? pos.second : pos.first;
+  }
+
   // Counts how many elements have key key.  For maps, it's either 0 or 1.
   size_type count(const key_type &key) const {
     pair<size_type, size_type> pos = find_position(key);
@@ -698,12 +856,22 @@ class sparse_hashtable {
 
   // Likewise, equal_range doesn't really make sense for us.  Oh well.
   pair<iterator,iterator> equal_range(const key_type& key) {
-    const iterator pos = find(key);      // either an iterator or end
-    return pair<iterator,iterator>(pos, pos);
+    iterator pos = find(key);      // either an iterator or end
+    if (pos == end()) {
+      return pair<iterator,iterator>(pos, pos);
+    } else {
+      const iterator startpos = pos++;
+      return pair<iterator,iterator>(startpos, pos);
+    }
   }
   pair<const_iterator,const_iterator> equal_range(const key_type& key) const {
-    const const_iterator pos = find(key);      // either an iterator or end
-    return pair<iterator,iterator>(pos, pos);
+    const_iterator pos = find(key);      // either an iterator or end
+    if (pos == end()) {
+      return pair<const_iterator,const_iterator>(pos, pos);
+    } else {
+      const const_iterator startpos = pos++;
+      return pair<const_iterator,const_iterator>(startpos, pos);
+    }
   }
 
 
@@ -711,6 +879,8 @@ class sparse_hashtable {
  private:
   // If you know *this is big enough to hold obj, use this routine
   pair<iterator, bool> insert_noresize(const value_type& obj) {
+    // First, double-check we're not inserting delkey
+    assert(!use_deleted || !equals(get_key(obj), delkey));
     const pair<size_type,size_type> pos = find_position(get_key(obj));
     if ( pos.first != ILLEGAL_BUCKET) {      // object was already there
       return pair<iterator,bool>(iterator(this, table.get_iter(pos.first),
@@ -732,7 +902,7 @@ class sparse_hashtable {
  public:
   // This is the normal insert routine, used by the outside world
   pair<iterator, bool> insert(const value_type& obj) {
-    resize_delta(1);                      // adding an object, grow if need be
+    resize_delta(1);  // adding an object, grow if need be
     return insert_noresize(obj);
   }
 
@@ -764,6 +934,8 @@ class sparse_hashtable {
 
   // DELETION ROUTINES
   size_type erase(const key_type& key) {
+    // First, double-check we're not erasing delkey.
+    assert(!use_deleted || !equals(key, delkey));
     const_iterator pos = find(key);   // shrug: shouldn't need to be const
     if ( pos != end() ) {
       assert(!test_deleted(pos));  // or find() shouldn't have returned it
@@ -776,10 +948,8 @@ class sparse_hashtable {
     }
   }
 
-  // This is really evil: really it should be iterator, not const_iterator.
-  // But...the only reason keys are const is to allow lookup.
-  // Since that's a moot issue for deleted keys, we allow const_iterators
-  void erase(const_iterator pos) {
+  // We return the iterator past the deleted item.
+  void erase(iterator pos) {
     if ( pos == end() ) return;    // sanity check
     if ( set_deleted(pos) ) {      // true if object has been newly deleted
       ++num_deleted;
@@ -787,6 +957,26 @@ class sparse_hashtable {
     }
   }
 
+  void erase(iterator f, iterator l) {
+    for ( ; f != l; ++f) {
+      if ( set_deleted(f)  )       // should always be true
+        ++num_deleted;
+    }
+    consider_shrink = true;        // will think about shrink after next insert
+  }
+
+  // We allow you to erase a const_iterator just like we allow you to
+  // erase an iterator.  This is in parallel to 'delete': you can delete
+  // a const pointer just like a non-const pointer.  The logic is that
+  // you can't use the object after it's erased anyway, so it doesn't matter
+  // if it's const or not.
+  void erase(const_iterator pos) {
+    if ( pos == end() ) return;    // sanity check
+    if ( set_deleted(pos) ) {      // true if object has been newly deleted
+      ++num_deleted;
+      consider_shrink = true;      // will think about shrink after next insert
+    }
+  }
   void erase(const_iterator f, const_iterator l) {
     for ( ; f != l; ++f) {
       if ( set_deleted(f)  )       // should always be true
@@ -840,44 +1030,51 @@ class sparse_hashtable {
   hasher hash;                      // required by hashed_associative_container
   key_equal equals;
   ExtractKey get_key;
+  SetKey set_key;
   size_type num_deleted;        // how many occupied buckets are marked deleted
-  bool use_deleted;                          // false until delval has been set
-  value_type delval;                         // which key marks deleted entries
+  bool use_deleted;                          // false until delkey has been set
+  // TODO(csilvers): make a pointer, and get rid of use_deleted (benchmark!)
+  key_type delkey;                           // which key marks deleted entries
+  float enlarge_resize_percent;                       // how full before resize
+  float shrink_resize_percent;                       // how empty before resize
+  size_type shrink_threshold;           // table.size() * shrink_resize_percent
+  size_type enlarge_threshold;         // table.size() * enlarge_resize_percent
   sparsetable<value_type> table;      // holds num_buckets and num_elements too
-  size_type shrink_threshold;                    // table.size() * HT_EMPTY_FLT
-  size_type enlarge_threshold;               // table.size() * HT_OCCUPANCY_FLT
   bool consider_shrink;   // true if we should try to shrink before next insert
+  int num_ht_copies;        // a statistics counter incremented every Copy/Move
 
   void reset_thresholds() {
-    enlarge_threshold = static_cast<size_type>(table.size()*HT_OCCUPANCY_FLT);
-    shrink_threshold = static_cast<size_type>(table.size()*HT_EMPTY_FLT);
+    enlarge_threshold = static_cast<size_type>(table.size()
+                                               * enlarge_resize_percent);
+    shrink_threshold = static_cast<size_type>(table.size()
+                                              * shrink_resize_percent);
     consider_shrink = false;   // whatever caused us to reset already considered
   }
 };
 
 // We need a global swap as well
-template <class V, class K, class HF, class ExK, class EqK, class A>
-inline void swap(sparse_hashtable<V,K,HF,ExK,EqK,A> &x,
-                 sparse_hashtable<V,K,HF,ExK,EqK,A> &y) {
+template <class V, class K, class HF, class ExK, class SetK, class EqK, class A>
+inline void swap(sparse_hashtable<V,K,HF,ExK,SetK,EqK,A> &x,
+                 sparse_hashtable<V,K,HF,ExK,SetK,EqK,A> &y) {
   x.swap(y);
 }
 
 #undef JUMP_
 
-template <class V, class K, class HF, class ExK, class EqK, class A>
-const typename sparse_hashtable<V,K,HF,ExK,EqK,A>::size_type
-  sparse_hashtable<V,K,HF,ExK,EqK,A>::ILLEGAL_BUCKET;
+template <class V, class K, class HF, class ExK, class SetK, class EqK, class A>
+const typename sparse_hashtable<V,K,HF,ExK,SetK,EqK,A>::size_type
+  sparse_hashtable<V,K,HF,ExK,SetK,EqK,A>::ILLEGAL_BUCKET;
 
 // How full we let the table get before we resize.  Knuth says .8 is
 // good -- higher causes us to probe too much, though saves memory
-template <class V, class K, class HF, class ExK, class EqK, class A>
-const float sparse_hashtable<V,K,HF,ExK,EqK,A>::HT_OCCUPANCY_FLT = 0.8f;
+template <class V, class K, class HF, class ExK, class SetK, class EqK, class A>
+const float sparse_hashtable<V,K,HF,ExK,SetK,EqK,A>::HT_OCCUPANCY_FLT = 0.8f;
 
 // How empty we let the table get before we resize lower.
 // It should be less than OCCUPANCY_FLT / 2 or we thrash resizing
-template <class V, class K, class HF, class ExK, class EqK, class A>
-const float sparse_hashtable<V,K,HF,ExK,EqK,A>::HT_EMPTY_FLT = 0.4f *
-sparse_hashtable<V,K,HF,ExK,EqK,A>::HT_OCCUPANCY_FLT;
+template <class V, class K, class HF, class ExK, class SetK, class EqK, class A>
+const float sparse_hashtable<V,K,HF,ExK,SetK,EqK,A>::HT_EMPTY_FLT = 0.4f *
+sparse_hashtable<V,K,HF,ExK,SetK,EqK,A>::HT_OCCUPANCY_FLT;
 
 _END_GOOGLE_NAMESPACE_
 
