@@ -56,6 +56,7 @@
 #include <iostream>
 #include <iomanip>             // for setprecision()
 #include <string>
+#include <stdexcept>           // for std::length_error
 #include HASH_FUN_H            // defined in config.h
 #include <google/type_traits.h>
 #include <google/dense_hash_map>
@@ -99,7 +100,7 @@ using STL_NAMESPACE::ostream;
 #define CHECK_LT(a, b)  CHECK((a) < (b))
 #define CHECK_GE(a, b)  CHECK((a) >= (b))
 
-#ifndef WIN32   // windows defines its own version
+#ifndef _MSC_VER   // windows defines its own version
 static string TmpFile(const char* basename) {
   return string("/tmp/") + basename;
 }
@@ -137,13 +138,6 @@ const char *default_dict[] = {"Aarhus\n",
                               "Zurich\n",
 };
 
-// Apparently identity is not stl-standard, so we define our own
-template<class Value>
-struct Identity {
-  Value& operator()(Value& v) const { return v; }
-  const Value& operator()(const Value& v) const { return v; }
-};
-
 // Likewise, it's not standard to hash a string pre-tr1.  Luckily, it is a char*
 #ifdef HAVE_UNORDERED_MAP
 typedef SPARSEHASH_HASH<string> StrHash;
@@ -151,6 +145,12 @@ struct CharStarHash {
   size_t operator()(const char* s) const {
     return StrHash()(string(s));
   }
+  // These are used by MSVC:
+  bool operator()(const char* a, const char* b) const {
+    return strcmp(a, b) < 0;
+  }
+  static const size_t bucket_size = 4;    // These are required by MSVC
+  static const size_t min_buckets = 8;    // 4 and 8 are the defaults
 };
 #else
 typedef SPARSEHASH_HASH<const char*> CharStarHash;
@@ -158,6 +158,12 @@ struct StrHash {
   size_t operator()(const string& s) const {
     return SPARSEHASH_HASH<const char*>()(s.c_str());
   }
+  // These are used by MSVC:
+  bool operator()(const string& a, const string& b) const {
+    return a < b;
+  }
+  static const size_t bucket_size = 4;    // These are required by MSVC
+  static const size_t min_buckets = 8;    // 4 and 8 are the defaults
 };
 #endif
 
@@ -482,6 +488,7 @@ void test_int() {
 
   CHECK(z.size() == 10);
   z.set_deleted_key(1010101010);      // an unused value
+  CHECK(z.deleted_key() == 1010101010);
   z.erase(11111);
   CHECK(z.size() == 9);
   insert(&z, 11111);                  // should retake deleted value
@@ -523,6 +530,7 @@ void test_int() {
   for ( typename htint::const_iterator it = z.begin(); it != z.end(); ++it )
     LOGF << "y: " << get_int_item(*it) << "\n";
   z.set_deleted_key(1010101011);      // a different unused value
+  CHECK(z.deleted_key() == 1010101011);
   for ( typename htint::const_iterator it = z.begin(); it != z.end(); ++it )
     LOGF << "y: " << get_int_item(*it) << "\n";
   LOGF << "That's " << z.size() << " elements\n";
@@ -594,12 +602,12 @@ void test_charptr(bool read_write) {
 
   // Save the hashtable.
   if (read_write) {
-    const string file_string = TmpFile("#hashtable_unittest_dicthash");
+    const string file_string = TmpFile(".hashtable_unittest_dicthash");
     const char* file = file_string.c_str();
     FILE *fp = fopen(file, "wb");
     if ( fp == NULL ) {
       // maybe we can't write to /tmp/.  Try the current directory
-      file = "#hashtable_unittest_dicthash";
+      file = ".hashtable_unittest_dicthash";
       fp = fopen(file, "wb");
     }
     if ( fp == NULL ) {
@@ -728,12 +736,12 @@ void test_string(bool read_write) {
 
   // Save the hashtable.
   if (read_write) {
-    const string file_string = TmpFile("#hashtable_unittest_dicthash_str");
+    const string file_string = TmpFile(".hashtable_unittest_dicthash_str");
     const char* file = file_string.c_str();
     FILE *fp = fopen(file, "wb");
     if ( fp == NULL ) {
       // maybe we can't write to /tmp/.  Try the current directory
-      file = "#hashtable_unittest_dicthash_str";
+      file = ".hashtable_unittest_dicthash_str";
       fp = fopen(file, "wb");
     }
     if ( fp == NULL ) {
@@ -1004,11 +1012,20 @@ static void TestMaps() {
   TestMap<dense_hash_map>();
 }
 
+static void TestCopyConstructor() {
+  {  // Copy an empty table with no empty key set.
+    dense_hash_map<int, string> table1;
+    dense_hash_map<int, string> table2(table1);
+    CHECK_EQ(32, table2.bucket_count());  // default number of buckets.
+  }
+}
+
 static void TestOperatorEquals() {
   {
     dense_hash_set<int> sa, sb;
     sa.set_empty_key(-1);
     sb.set_empty_key(-1);
+    CHECK(sa.empty_key() == -1);
     sa.set_deleted_key(-2);
     sb.set_deleted_key(-2);
     CHECK(sa == sb);
@@ -1027,6 +1044,7 @@ static void TestOperatorEquals() {
     dense_hash_map<int, string> sa, sb;
     sa.set_empty_key(-1);
     sb.set_empty_key(-1);
+    CHECK(sa.empty_key() == -1);
     sa.set_deleted_key(-2);
     sb.set_deleted_key(-2);
     CHECK(sa == sb);
@@ -1043,6 +1061,54 @@ static void TestOperatorEquals() {
     sb.erase(1);
     CHECK(sa != sb);
   }
+  {  // Copy table with a different empty key.
+    dense_hash_map<string, string, StrHash> table1;
+    table1.set_empty_key("key1");
+    CHECK(table1.empty_key() == "key1");
+    dense_hash_map<string, string, StrHash> table2;
+    table2.set_empty_key("key2");
+    table1.insert(make_pair("key", "value"));
+    table2.insert(make_pair("a", "b"));
+    table1 = table2;
+    CHECK_EQ("b", table1["a"]);
+    CHECK_EQ(1, table1.size());
+  }
+  {  // Assign to a map without an empty key.
+    dense_hash_map<string, string, StrHash> table1;
+    dense_hash_map<string, string, StrHash> table2;
+    table2.set_empty_key("key2");
+    CHECK(table2.empty_key() == "key2");
+    table2.insert(make_pair("key", "value"));
+    table1 = table2;
+    CHECK_EQ("value", table1["key"]);
+  }
+  {  // Copy a map without an empty key.
+    dense_hash_map<string, string, StrHash> table1;
+    dense_hash_map<string, string, StrHash> table2;
+    table1 = table2;
+    CHECK_EQ(0, table1.size());
+    table1.set_empty_key("key1");
+    CHECK(table1.empty_key() == "key1");
+    table1.insert(make_pair("key", "value"));
+    table1 = table2;
+    CHECK_EQ(0, table1.size());
+    table1.set_empty_key("key1");
+    table1.insert(make_pair("key", "value"));
+    CHECK_EQ("value", table1["key"]);
+  }
+  {
+    sparse_hash_map<string, string, StrHash> table1;
+    table1.set_deleted_key("key1");
+    table1.insert(make_pair("key", "value"));
+
+    sparse_hash_map<string, string, StrHash> table2;
+    table2.set_deleted_key("key");
+    table2 = table1;
+
+    CHECK_EQ(1, table2.size());
+    table2.erase("key");
+    CHECK(table2.empty());
+  }
 }
 
 // Test the interface for setting the resize parameters in a
@@ -1109,6 +1175,60 @@ static void TestResizingParameters() {
   }
 }
 
+// This tests for a problem we had where we could repeatedly "resize"
+// a hashtable to the same size it was before, on every insert.
+template<class HT>
+static void TestHashtableResizing() {
+  HT ht;
+  ht.set_deleted_key(-1);
+  set_empty_key(&ht, -2);
+
+  const int kSize = 1<<10;       // Pick any power of 2
+  const float kResize = 0.8f;    // anything between 0.5 and 1 is fine.
+  const int kThreshold = kSize * kResize - 1;
+  ht.set_resizing_parameters(0, kResize);
+
+  // Get right up to the resizing threshold.
+  for (int i = 0; i <= kThreshold; i++) {
+    ht.insert(i);
+  }
+  // The bucket count should equal kSize.
+  CHECK_EQ(ht.bucket_count(), kSize);
+
+  // Now start doing erase+insert pairs.  This should cause us to
+  // copy the hashtable at most once.
+  const int pre_copies = ht.num_table_copies();
+  for (int i = 0; i < kSize; i++) {
+    if (i % 100 == 0)
+      LOGF << "erase/insert: " << i
+           << " buckets: " << ht.bucket_count()
+           << " size: " << ht.size();
+    ht.erase(kThreshold);
+    ht.insert(kThreshold);
+  }
+  CHECK_LT(ht.num_table_copies(), pre_copies + 2);
+
+  // Now create a hashtable where we go right to the threshold, then
+  // delete everything and do one insert.  Even though our hashtable
+  // is now tiny, we should still have at least kSize buckets, because
+  // our shrink threshhold is 0.
+  HT ht2;
+  ht2.set_deleted_key(-1);
+  set_empty_key(&ht2, -2);
+  ht2.set_resizing_parameters(0, kResize);
+  CHECK_LT(ht2.bucket_count(), kSize);
+  for (int i = 0; i <= kThreshold; i++) {
+    ht2.insert(i);
+  }
+  CHECK_EQ(ht2.bucket_count(), kSize);
+  for (int i = 0; i <= kThreshold; i++) {
+    ht2.erase(i);
+    CHECK_EQ(ht2.bucket_count(), kSize);
+  }
+  ht2.insert(kThreshold+1);
+  CHECK_GE(ht2.bucket_count(), kSize);
+}
+
 // Tests the some of the tr1-inspired API features.
 template<class HS>
 static void TestTR1API() {
@@ -1275,6 +1395,26 @@ void TestMemoryManagement() {
   CHECK_EQ(2, MemUsingKey::net_allocations());  // for deleted+empty_key
 }
 
+void TestHugeResize() {
+  try {
+    dense_hash_map<int, int> ht;
+    ht.resize(static_cast<size_t>(-1));
+    LOGF << "dense_hash_map resize should have failed";
+    abort();
+  } catch (const std::length_error&) {
+    // Good, the resize failed.
+  }
+
+  try {
+    sparse_hash_map<int, int> ht;
+    ht.resize(static_cast<size_t>(-1));
+    LOGF << "sparse_hash_map resize should have failed";
+    abort();
+  } catch (const std::length_error&) {
+    // Good, the resize failed.
+  }
+}
+
 template<class Key>
 struct SetKey {
   void operator()(Key* key, const Key& new_key) const {
@@ -1282,7 +1422,15 @@ struct SetKey {
   }
 };
 
+template<class Value>
+struct Identity {
+  Value& operator()(Value& v) const { return v; }
+  const Value& operator()(const Value& v) const { return v; }
+};
+
+
 int main(int argc, char **argv) {
+  TestCopyConstructor();
   TestOperatorEquals();
 
   // SPARSEHASH_HASH is defined in sparseconfig.h.  It resolves to the
@@ -1295,8 +1443,7 @@ int main(int argc, char **argv) {
                        Identity<char *>, SetKey<char *>, strcmp_fnc,
 		       allocator<char *> >,
        dense_hashtable<string, string, StrHash,
-                       Identity<string>, SetKey<string>,
-		       equal_to<string>,
+                       Identity<string>, SetKey<string>, equal_to<string>,
 		       allocator<string> >,
        dense_hashtable<int, int, SPARSEHASH_HASH<int>,
                        Identity<int>, SetKey<int>, equal_to<int>,
@@ -1312,6 +1459,14 @@ int main(int argc, char **argv) {
   TestResizingParameters<dense_hash_set<int>, true>();    // use tr1 API
   TestResizingParameters<dense_hash_set<int>, false>();   // use older API
 
+  TestHashtableResizing<dense_hashtable<int, int, SPARSEHASH_HASH<int>,
+                           Identity<int>, SetKey<int>, equal_to<int>,
+                           allocator<int> > >();
+  TestHashtableResizing<sparse_hashtable<int, int, SPARSEHASH_HASH<int>,
+                           Identity<int>, SetKey<int>, equal_to<int>,
+                           allocator<int> > >();
+
+
   // Now try with hash_map, which differs only in insert()
   LOGF << "\n\nTEST WITH DENSE_HASH_MAP\n\n";
   test<dense_hash_map<char *, int, CharStarHash, strcmp_fnc>,
@@ -1321,15 +1476,14 @@ int main(int argc, char **argv) {
   // First try with the low-level hashtable interface
   LOGF << "\n\nTEST WITH SPARSE_HASHTABLE\n\n";
   test<sparse_hashtable<char *, char *, CharStarHash,
-                       Identity<char *>, SetKey<char *>, strcmp_fnc,
-		       allocator<char *> >,
+                        Identity<char *>, SetKey<char *>, strcmp_fnc,
+                        allocator<char *> >,
        sparse_hashtable<string, string, StrHash,
-                       Identity<string>, SetKey<string>, equal_to<string>,
-		       allocator<string> >,
+                        Identity<string>, SetKey<string>, equal_to<string>,
+                        allocator<string> >,
        sparse_hashtable<int, int, SPARSEHASH_HASH<int>,
-                       Identity<int>, SetKey<int>, equal_to<int>,
-		       allocator<int> > >(
-                         true);
+                        Identity<int>, SetKey<int>, equal_to<int>,
+                        allocator<int> > >(true);
 
   // Now try with hash_set, which should be equivalent
   LOGF << "\n\nTEST WITH SPARSE_HASH_SET\n\n";
@@ -1370,6 +1524,8 @@ int main(int argc, char **argv) {
   LOGF << "\n\nTesting memory management\n";
   TestMemoryManagement();
 
+  TestHugeResize();
+
   LOGF << "\nAll tests pass.\n";
   return 0;
 }
